import os

from Crypto.Cipher import AES, ChaCha20_Poly1305
from Crypto.Hash import SHA256
from Crypto.Protocol.KDF import HKDF

from ..config import ECIES_CONFIG

AES_CIPHER_MODE = AES.MODE_GCM
AEAD_TAG_LENGTH = 16
XCHACHA20_NONCE_LENGTH = 24


def sym_encrypt(key: bytes, plain_text: bytes) -> bytes:
    """
    Symmetric encryption. AES-256-GCM or XChaCha20-Poly1305.

    Nonce may be 12/16 bytes on AES, 24 bytes on XChaCha. Default is AES-256-GCM with 16 bytes nonce.

    Parameters
    ----------
    key: bytes
        Symmetric encryption session key, which derived from two secp256k1 keys
    plain_text: bytes
        Plain text to encrypt

    Returns
    -------
    bytes
        nonce + tag(16 bytes) + encrypted data
    """
    algorithm = ECIES_CONFIG.symmetric_algorithm
    if algorithm == "aes-256-gcm":
        nonce_length = ECIES_CONFIG.symmetric_nonce_length
        nonce = os.urandom(nonce_length)
        cipher = AES.new(key, AES_CIPHER_MODE, nonce)
    elif algorithm == "xchacha20":
        nonce = os.urandom(XCHACHA20_NONCE_LENGTH)
        cipher = ChaCha20_Poly1305.new(key=key, nonce=nonce)  # type:ignore
    else:
        raise NotImplementedError

    encrypted, tag = cipher.encrypt_and_digest(plain_text)
    cipher_text = bytearray()
    cipher_text.extend(nonce)
    cipher_text.extend(tag)
    cipher_text.extend(encrypted)
    return bytes(cipher_text)


def sym_decrypt(key: bytes, cipher_text: bytes) -> bytes:
    """
    AES-GCM decryption. AES-256-GCM or XChaCha20-Poly1305.

    Parameters
    ----------
    key: bytes
        Symmetric encryption session key, which derived from two secp256k1 keys
    cipher_text: bytes
        Encrypted text:
            nonce + tag(16 bytes) + encrypted data

    Returns
    -------
    bytes
        Plain text

    >>> from coincurve.utils import get_valid_secret
    >>> data = b'this is test data'
    >>> key = get_valid_secret()
    >>> sym_decrypt(key, sym_encrypt(key, data)) == data
    True
    >>> import os
    >>> key = os.urandom(32)
    >>> sym_decrypt(key, sym_encrypt(key, data)) == data
    True
    """

    # NOTE
    # pycryptodome's aes gcm takes nonce as iv
    # but actually nonce (12 bytes) should be used to generate iv (16 bytes) and iv should be sequential
    # See https://crypto.stackexchange.com/a/71219
    # You can configure to use 12 bytes nonce by setting `ECIES_CONFIG.symmetric_nonce_length = 12`
    # If it's 12 bytes, the nonce can be incremented by 1 for each encryption
    # If it's 16 bytes, the nonce will be used to hash, so it's meaningless to increment

    algorithm = ECIES_CONFIG.symmetric_algorithm
    if algorithm == "aes-256-gcm":
        nonce_length = ECIES_CONFIG.symmetric_nonce_length
        nonce_tag_length = nonce_length + AEAD_TAG_LENGTH
        nonce = cipher_text[:nonce_length]
        tag = cipher_text[nonce_length:nonce_tag_length]
        ciphered_data = cipher_text[nonce_tag_length:]
        cipher = AES.new(key, AES_CIPHER_MODE, nonce)
    elif algorithm == "xchacha20":
        nonce_tag_length = XCHACHA20_NONCE_LENGTH + AEAD_TAG_LENGTH
        nonce = cipher_text[:XCHACHA20_NONCE_LENGTH]
        tag = cipher_text[XCHACHA20_NONCE_LENGTH:nonce_tag_length]
        ciphered_data = cipher_text[nonce_tag_length:]
        cipher = ChaCha20_Poly1305.new(key=key, nonce=nonce)  # type:ignore
    else:
        raise NotImplementedError
    return cipher.decrypt_and_verify(ciphered_data, tag)


def derive_key(master: bytes) -> bytes:
    derived = HKDF(master, 32, b"", SHA256)
    return derived  # type: ignore



from .elliptic import (
    decapsulate,
    encapsulate,
    generate_eth_key,
    generate_key,
    hex2pk,
    hex2sk,
)
from .hex import decode_hex, sha256
from .symmetric import sym_decrypt, sym_encrypt

__all__ = [
    "sha256",
    "decode_hex",
    "sym_encrypt",
    "sym_decrypt",
    "generate_key",
    "generate_eth_key",
    "hex2sk",
    "hex2pk",
    "decapsulate",
    "encapsulate",
]



from dataclasses import dataclass
from typing import Literal

SymmetricAlgorithm = Literal["aes-256-gcm", "xchacha20"]
NonceLength = Literal[12, 16]  # only for aes-256-gcm, xchacha20 will always be 24

COMPRESSED_PUBLIC_KEY_SIZE = 33
UNCOMPRESSED_PUBLIC_KEY_SIZE = 65


@dataclass()
class Config:
    is_ephemeral_key_compressed: bool = False
    is_hkdf_key_compressed: bool = False
    symmetric_algorithm: SymmetricAlgorithm = "aes-256-gcm"
    symmetric_nonce_length: NonceLength = 16

    @property
    def ephemeral_key_size(self):
        return (
            COMPRESSED_PUBLIC_KEY_SIZE
            if self.is_ephemeral_key_compressed
            else UNCOMPRESSED_PUBLIC_KEY_SIZE
        )


ECIES_CONFIG = Config()



from typing import Union

from coincurve import PrivateKey, PublicKey

from .config import ECIES_CONFIG
from .utils import (
    decapsulate,
    encapsulate,
    generate_key,
    hex2pk,
    hex2sk,
    sym_decrypt,
    sym_encrypt,
)

__all__ = ["encrypt", "decrypt", "ECIES_CONFIG"]


def encrypt(receiver_pk: Union[str, bytes], msg: bytes) -> bytes:
    """
    Encrypt with receiver's secp256k1 public key

    Parameters
    ----------
    receiver_pk: Union[str, bytes]
        Receiver's public key (hex str or bytes)
    msg: bytes
        Data to encrypt

    Returns
    -------
    bytes
        Encrypted data
    """
    if isinstance(receiver_pk, str):
        pk = hex2pk(receiver_pk)
    elif isinstance(receiver_pk, bytes):
        pk = PublicKey(receiver_pk)
    else:
        raise TypeError("Invalid public key type")

    ephemeral_sk = generate_key()
    ephemeral_pk = ephemeral_sk.public_key.format(
        ECIES_CONFIG.is_ephemeral_key_compressed
    )

    sym_key = encapsulate(ephemeral_sk, pk)
    encrypted = sym_encrypt(sym_key, msg)
    return ephemeral_pk + encrypted


def decrypt(receiver_sk: Union[str, bytes], msg: bytes) -> bytes:
    """
    Decrypt with receiver's secp256k1 private key

    Parameters
    ----------
    receiver_sk: Union[str, bytes]
        Receiver's private key (hex str or bytes)
    msg: bytes
        Data to decrypt

    Returns
    -------
    bytes
        Plain text
    """
    if isinstance(receiver_sk, str):
        sk = hex2sk(receiver_sk)
    elif isinstance(receiver_sk, bytes):
        sk = PrivateKey(receiver_sk)
    else:
        raise TypeError("Invalid secret key type")

    key_size = ECIES_CONFIG.ephemeral_key_size
    ephemeral_pk, encrypted = PublicKey(msg[0:key_size]), msg[key_size:]

    sym_key = decapsulate(ephemeral_pk, sk)
    return sym_decrypt(sym_key, encrypted)



from coincurve import PrivateKey, PublicKey
from coincurve.utils import get_valid_secret
from eth_keys import keys

from ..config import ECIES_CONFIG
from .hex import decode_hex
from .symmetric import derive_key


def generate_key() -> PrivateKey:
    """
    Generate a random coincurve.PrivateKey`

    Returns
    -------
    coincurve.PrivateKey
        A secp256k1 key

    >>> k = generate_key()
    """
    return PrivateKey(get_valid_secret())


def generate_eth_key():
    """
    Generate a random `eth_keys.keys.PrivateKey`

    Returns
    -------
    eth_keys.keys.PrivateKey
        An ethereum key

    >>> k = generate_eth_key()
    """
    return keys.PrivateKey(get_valid_secret())


def hex2pk(pk_hex: str) -> PublicKey:
    """
    Convert ethereum hex to `coincurve.PublicKey`
    The hex should be 65 bytes (uncompressed) or 33 bytes (compressed), but ethereum public key has 64 bytes.
    `0x04` will be appended if it's an ethereum public key.

    Parameters
    ----------
    pk_hex: str
        Public key hex string

    Returns
    -------
    coincurve.PublicKey
        A secp256k1 public key

    >>> from ecies.utils import sha256
    >>> data = b'0' * 32
    >>> data_hash = sha256(data)
    >>> eth_sk = generate_eth_key()
    >>> cc_sk = hex2sk(eth_sk.to_hex())
    >>> eth_sk.sign_msg_hash(data_hash).to_bytes() == cc_sk.sign_recoverable(data)
    True
    >>> pk_hex = eth_sk.public_key.to_hex()
    >>> computed_pk = hex2pk(pk_hex)
    >>> computed_pk == cc_sk.public_key
    True
    """
    uncompressed = decode_hex(pk_hex)
    if len(uncompressed) == 64:  # eth public key format
        uncompressed = b"\x04" + uncompressed

    return PublicKey(uncompressed)


def hex2sk(sk_hex: str) -> PrivateKey:
    """
    Convert ethereum hex to `coincurve.PrivateKey`

    Parameters
    ----------
    sk_hex: str
        Private key hex string

    Returns
    -------
    coincurve.PrivateKey
        A secp256k1 private key

    >>> k = generate_eth_key()
    >>> sk_hex = k.to_hex()
    >>> pk_hex = k.public_key.to_hex()
    >>> computed_sk = hex2sk(sk_hex)
    >>> computed_sk.to_int() == int(k.to_hex(), 16)
    True
    """
    return PrivateKey(decode_hex(sk_hex))


# private below
def encapsulate(private_key: PrivateKey, peer_public_key: PublicKey) -> bytes:
    is_compressed = ECIES_CONFIG.is_hkdf_key_compressed
    shared_point = peer_public_key.multiply(private_key.secret)
    master = private_key.public_key.format(is_compressed) + shared_point.format(
        is_compressed
    )
    return derive_key(master)


def decapsulate(public_key: PublicKey, peer_private_key: PrivateKey) -> bytes:
    is_compressed = ECIES_CONFIG.is_hkdf_key_compressed
    shared_point = public_key.multiply(peer_private_key.secret)
    master = public_key.format(is_compressed) + shared_point.format(is_compressed)
    return derive_key(master)



import codecs
import hashlib


def sha256(msg: bytes) -> bytes:
    """
    Calculate sha256 hash.

    Parameters
    ----------
    msg: bytes
        message to hash

    Returns
    -------
    bytes
        sha256 hash in bytes

    >>> sha256(b'0'*16).hex()[:8] == 'fcdb4b42'
    True
    """
    return hashlib.sha256(msg).digest()


def decode_hex(s: str) -> bytes:
    """
    Decode hex string to bytes. `0x` prefix is optional.

    Parameters
    ----------
    s: str
        hex string

    Returns
    -------
    bytes
        decoded bytes

    >>> decode_hex('0x7468697320697320612074657374') == b'this is a test'
    True
    """
    return codecs.decode(remove_0x(s), "hex")


# private below
def remove_0x(s: str) -> str:
    if s.startswith("0x") or s.startswith("0X"):
        return s[2:]
    return s
